// ========================================================================================= //
//                                                                                           //
//                            (pipe)                       //
//                                    2017-03-27                                       //
//                                                                                           //
//                                       HiAsm                                         //
//                                     www.hiasm.com                                         //
//                                                                                           //
//        : Netspirit                                                                   //
//        E-mail: netspirit@meta.ua                                                          //
//                                                                                           //
//                                                       //
//                                                                                           //
//        1)    " ".                      //
//              .                                                //
//        2)          ,                     //
//                ,                  //
//            .                                                                     //
//        3)                             //
//           (  )   .                            //
//                                //
//                .                                //
//        4)                          //
//                                   //
//                  .                   //
//        5)        (, )         //
//                .                                //
//                                                                                           //
//                        ,                              //
//                                                                  //
//                                                                                           //
//                                           //
//                                                                                           //
// ========================================================================================= //



unit NSPipes;

interface

uses
  Windows, Kol, NSThreads;


type

  //   
  TBuffer = class
    private
      FSize: Integer; //   FBuffer
      FBuffer: Pointer;
      procedure SetSize(Value: Integer);
      
    public
      Len: Integer; //  
      property Size: Integer read FSize write SetSize;
      property Buffer: Pointer read FBuffer;
      
      destructor Destroy; override;
      function GetString: string; //   FBuffer     Len
  end;
  
  TBasicPipe = class(TSharedObject)
    private
      FEvWrite: THandle;
      
    protected
      FErrorCode: Integer;
      FPipeHandle: THandle;
      
      FAbortWrite: Boolean;
      FWriteStarted: Boolean;
      
    public
      WriteTimeout: Integer; //      FILE_FLAG_OVERLAPPED
      
      ReadBufferSize: Integer;
      WriteBufferSize: Integer;
      
      property ErrorCode: Integer read FErrorCode;
      property Handle: THandle read FPipeHandle;
      property Writing: Boolean read FWriteStarted;
      
      constructor Create(Handle: THandle);
      destructor Destroy; override;

      procedure AbortWrite;
      
      //        FILE_FLAG_OVERLAPPED
      //   ,    
      //     
      procedure Close;
      
      // ==========  / ========== //
      //  Pipe,   FILE_FLAG_OVERLAPPED   -
      //      
      // (  )
      
      //  .    .
      //     FILE_FLAG_OVERLAPPED
      // RESULT_ERROR - ,  ErrorCode
      function Read(Buf: Pointer; BufSize: Integer): Integer;
      
      //  .    .
      // RESULT_TIMEOUT -   (  FILE_FLAG_OVERLAPPED)
      // RESULT_ERROR - ,  ErrorCode
      function Write(Buf: Pointer; BufSize: Integer): Integer;
  end;
  
  TCustomPipe = class(TBasicPipe)
    private
      FReadingThread: TThread;
      FCSDisconnect: TRTLCriticalSection;
      
      function _ExecuteReceiving(T: TCustomThread): Integer;
      procedure _SyncDisconnect(Param: Pointer);
      procedure _SyncReceive(Param: Pointer);
      
    protected
      procedure AfterClose(Reason: Integer); virtual;
      procedure BeforeClose(Reason: Integer); virtual;
      procedure DataReceived(Buf: TBuffer); virtual;
      procedure InternalClose(Reason: Integer);
      
    public
      AsyncEvents: Boolean;
      
      constructor Create(Handle: THandle);
      destructor Destroy; override;
      
      procedure Disconnect; //  
      
      function Send(const S: string): Integer;
      procedure StartReceiving;
      procedure StopReceiving;
  end;
  
  // Pipe-  
  TCustomPipeClient = class(TCustomPipe)
    private
      FPipeMode: Integer;
      FPipeName: string;
      
    public
      ConnectTimeout: Integer;
      property PipeMode: Integer read FPipeMode; // GENERIC_READ, GENERIC_WRITE, GENERIC_READ or GENERIC_WRITE
      property PipeName: string read FPipeName;
      
      constructor Create;
      
      //    
      // Name -  
      // Mode -   (GENERIC_READ / GENERIC_WRITE)
      function Connect(Name: string; Mode: Integer): Boolean;
  end;
  
  //   
  TBasicPipeServer = class
    private
      FAcceptThread: TThread;
      FPipeHandle: THandle; //    
      FPipeMode: Integer;
      FPipeName: string;

      function _ExecuteAccept(T: TCustomThread): Integer;
      procedure _SyncClientConnect(Param: Pointer); // Param  Handle
      procedure _SyncListeningError(Param: Pointer); // Param  ErrorCode
      function CreatePipeInstance: THandle;
      function GetIsListening: Boolean;
      
    protected
      procedure ConnectionAccepted(PipeHandle: THandle); virtual;
      procedure ListeninigFailed(ErrorCode: Integer); virtual; //     
      
    public
      AsyncEvents: Boolean;
      ReadBufferSize: Integer;
      WriteBufferSize: Integer;
      WriteTimeout: Integer;
      
      property IsListening: Boolean read GetIsListening;
      property PipeMode: Integer read FPipeMode; // PIPE_ACCESS_INBOUND, PIPE_ACCESS_OUTBOUND, PIPE_ACCESS_DUPLEX
      property PipeName: string read FPipeName;
      
      constructor Create;
      destructor Destroy; override;
      
      //    pipe    .
      //      -  
      // Name -  pipe, Mode -  
      function StartListeninig(Name: string; Mode: Integer): Integer; 
      procedure StopListeninig; //  
  end;
  
  TCustomPipeServer = class;
  
  // ,    pipe-
  TCustomPipeConnection = class(TCustomPipe)
    protected
      Server: TCustomPipeServer;
      procedure AfterClose(Reason: Integer); override; //   (!)
      procedure DataReceived(Buf: TBuffer); override;
      
    public
      
  end;
  
  // Pipe-  
  TCustomPipeServer = class(TBasicPipeServer)
    private
      FConnList: PList;
      FCSList: TRTLCriticalSection;
      FSendAllStarted: Boolean;
      FAbortSendAll: Boolean;

      function GetCount: Integer;
      
    protected
      //      .       .
      //   ,     .
      procedure AfterNewConnection(Conn: TCustomPipeConnection); virtual;
      
      //      .
      //        TCustomPipeConnection,
      //     TCustomPipeConnection.
      //  nil,   
      function BeforeNewConnection(PipeHandle: THandle): TCustomPipeConnection; virtual;
      
      procedure ConnectionAccepted(PipeHandle: THandle); override;
      procedure ConnectionAdd(Conn: TCustomPipeConnection);
      procedure ConnectionDelete(Conn: TCustomPipeConnection);
      procedure DataReceived(Conn: TCustomPipeConnection; Buf: TBuffer); virtual;
      procedure Disconnected(Conn: TCustomPipeConnection); virtual;
      function HoldConnectionIdx(I: Integer): TCustomPipeConnection;
      procedure SendResult(Conn: TCustomPipeConnection; ResultCode: Integer); virtual; // ResultCode = 0 -  ,   
      
    public
      property Count: Integer read GetCount; //   

      constructor Create;
      destructor Destroy; override;
      
      procedure AbortSend(Conn: TCustomPipeConnection);
      procedure AbortSendAll;
      
      procedure Close(Conn: TCustomPipeConnection); //  .   (!)
      procedure CloseAll; //   .   (!)
      procedure GetConnections(OutConn: PList); //    
      //   ,   -      ().
      function HoldConnection(Conn: TCustomPipeConnection): TCustomPipeConnection;
      
      procedure Send(Conn: TCustomPipeConnection; S: string);
      procedure SendAll(S: string);
      
      procedure StopServer; //      
  end;
  
  
  
  
  // ==============   ============== //
  
  
  //  Pipe-
  TPipeClient = class(TCustomPipeClient)
    protected
      procedure AfterClose(Reason: Integer); override;
      procedure DataReceived(Buf: TBuffer); override;
      
    public
      OnDisconnect: procedure (Sender: TCustomPipe; Reason: Integer) of object;
      OnReceive: procedure (Sender: TCustomPipe; Buf: TBuffer) of object;
  end;
  
  //  Pipe-
  TPipeServer = class(TCustomPipeServer)
    protected
      procedure AfterNewConnection(Conn: TCustomPipeConnection); override;
      function BeforeNewConnection(PipeHandle: THandle): TCustomPipeConnection; override;
      procedure DataReceived(Conn: TCustomPipeConnection; Buf: TBuffer); override;
      procedure Disconnected(Conn: TCustomPipeConnection); override;
      //procedure ListeninigFailed(ErrorCode: Integer); override;
      procedure SendResult(Conn: TCustomPipeConnection; ResultCode: Integer); override;
      
    public
      OnClientConnect: procedure (Conn: TCustomPipeConnection) of object;
      OnClientDisconnect: procedure (Conn: TCustomPipeConnection) of object;
      OnReceive: procedure (Conn: TCustomPipeConnection; Buf: TBuffer) of object;
      OnSend: procedure (Conn: TCustomPipeConnection) of object;
      OnErrorSend: procedure (Conn: TCustomPipeConnection; ErrorCode: Integer) of object;
  end;
  
  
  
  function CancelIo(hFile: THandle): BOOL; stdcall; external 'kernel32.dll' name 'CancelIo';

  //     
  // :
  // >= 0 -    
  // RESULT_TIMEOUT -  
  // RESULT_ERROR - , GetLastError()   
  //   Overlap    hEvent
  
  function WaitForOverlappedResult(Handle: Thandle; var Overlap: TOverlapped; Timeout: Integer): Integer;

const
  
  DISCONNECT_REASON_MANUAL     = 0;
  DISCONNECT_REASON_REMOTE     = 1;
  DISCONNECT_REASON_DESTROYING = 2;
  
  RESULT_ERROR = -2;
  RESULT_TIMEOUT = -1;
  
  FILE_FLAG_FIRST_PIPE_INSTANCE = $00080000;
  {$ifdef FPC_NEW}PIPE_UNLIMITED_INSTANCES = 255;{$endif}

implementation


const
  DEFAULT_BUFFER_SIZE = 8192;
  DEFAULT_WRITE_TIMEOUT = 5000;



function WaitForOverlappedResult(Handle: THandle; var Overlap: TOverlapped; Timeout: Integer): Integer;
var
  B: DWORD;
begin
  B := WaitForSingleObject(Overlap.hEvent, Timeout);
  
  case B of
    WAIT_OBJECT_0:
      begin
        if GetOverlappedResult(Handle, Overlap, B, False) then
          Result := B
        else
          Result := RESULT_ERROR;
      end;
    WAIT_TIMEOUT:
      Result := RESULT_TIMEOUT;
    else
      Result := RESULT_ERROR;
  end;
end;

// ====== TBuffer ====== //

destructor TBuffer.Destroy;
begin
  SetSize(0);
  inherited;
end;

procedure TBuffer.SetSize(Value: Integer);
begin
  if Value = FSize then Exit;
  
  if Value < 1 then
  begin
    if FBuffer <> nil then
    begin
      FreeMem(FBuffer, FSize);
      FBuffer := nil;
      FSize := 0;
    end;
  end
  else
  begin
    if FBuffer <> nil then FreeMem(FBuffer, FSize);
    FSize := Value;
    GetMem(FBuffer, Value);
  end;
end;


function TBuffer.GetString: string;
begin
  if (FSize > 0) and (Len > 0) then // TODO:  Len <= FSize
  begin
    SetLength(Result, Len);
    Move(FBuffer^, Result[1], Len);
  end;
end;

// ======================== //

// ====== TBasicPipe ====== //

constructor TBasicPipe.Create(Handle: THandle);
begin
  inherited Create;
  FPipeHandle := Handle;
  
  ReadBufferSize := DEFAULT_BUFFER_SIZE;
  WriteBufferSize := DEFAULT_BUFFER_SIZE;
  WriteTimeout := DEFAULT_WRITE_TIMEOUT;
end;

destructor TBasicPipe.Destroy;
begin
  Close;
  CloseHandle(FEvWrite);
  inherited;
end;

procedure TBasicPipe.AbortWrite;
begin
  FAbortWrite := True;
end;

procedure TBasicPipe.Close;
begin
  AbortWrite;
  if FPipeHandle <> 0 then
  begin
    //FlushFileBuffers(FPipeHandle);
    //DisconnectNamedPipe(FPipeHandle);
    CloseHandle(FPipeHandle);
    FPipeHandle := 0;
  end;
end;



function TBasicPipe.Read(Buf: Pointer; BufSize: Integer): Integer;
begin
  if ReadFile(FPipeHandle, Buf^, BufSize, DWORD(Result), nil) then
  begin
    FErrorCode := 0;
  end
  else
  begin
    FErrorCode := GetLastError();
    Result := RESULT_ERROR;
  end;
end;

function TBasicPipe.Write(Buf: Pointer; BufSize: Integer): Integer;
label
  finish;
var
  Err: Integer;
  BW, NeedW, TotalW: Integer;
  Overlap: TOverlapped;
begin
  Result := RESULT_ERROR;
  
  if FWriteStarted then
  begin
    FErrorCode := ERROR_BUSY;
    Exit;    
  end;
  
  TotalW := 0;
  
  
  FErrorCode := 0;
  FAbortWrite := False;
  FWriteStarted := True;
  
  
  if FEvWrite = 0 then FEvWrite := CreateEvent(nil, False{ManualReset}, False, nil);
  
  ZeroMemory(@Overlap, SizeOf(TOverlapped));
  Overlap.hEvent := FEvWrite;
  
  while BufSize > 0 do
  begin
    NeedW := WriteBufferSize;
    if NeedW > BufSize then NeedW := BufSize;
    
    //ZeroMemory(@Overlap, SizeOf(TOverlapped));
    //Overlap.hEvent := FEvWrite;
    
    if not WriteFile(FPipeHandle, Buf^, NeedW, DWORD(BW), @Overlap) then
    begin
      if FAbortWrite then
      begin
        FErrorCode := ERROR_OPERATION_ABORTED;
        goto finish;
      end;    
    
      Err := GetLastError();
      
      if Err = ERROR_IO_PENDING then
      begin
        BW := WaitForOverlappedResult(FPipeHandle, Overlap, WriteTimeout);

        if FAbortWrite then
        begin
          CancelIo(FPipeHandle);
          FErrorCode := ERROR_OPERATION_ABORTED;
          goto finish;
        end;
        
        if BW = RESULT_ERROR then // BW = RESULT_ERROR - 
        begin
          FErrorCode := GetLastError();
          goto finish;
        end;
          
        if BW = RESULT_TIMEOUT then // BW = RESULT_TIMEOUT - 
        begin
          CancelIo(FPipeHandle);
          Result := RESULT_TIMEOUT;
          goto finish;
        end;
      end
      else
      begin
        FErrorCode := Err;
        goto finish;
      end;
    end;
    
    Buf := Pointer(Integer(Buf) + BW);
    Dec(BufSize, BW);
    Inc(TotalW, BW);
  end;
  
  Result := TotalW;
 
  
finish:
  FWriteStarted := False;
end;

(*function TBasicPipe.Read(Buf: Pointer; BufSize: Integer): Integer;
var
  BR: DWORD;
begin
  if ReadFile(FPipeHandle, Buf^, BufSize, BR, nil) then
  begin
    FErrorCode := 0;
    Result := BR;
  end
  else
  begin
    FErrorCode := GetLastError();
    Result := -1;
  end;
end;

function TBasicPipe.Write(Buf: Pointer; BufSize: Integer): Integer;
var
  BW: DWORD;
begin
  if WriteFile(FPipeHandle, Buf^, BufSize, BW, nil) then
  begin
    FErrorCode := 0;
    Result := BW;
  end
  else
  begin
    FErrorCode := GetLastError();
    Result := -1;
  end;
end;*)


// ====== TCustomPipe ====== //

constructor TCustomPipe.Create(Handle: THandle);
begin
  inherited;
  InitializeCriticalSection(FCSDisconnect);
end;

destructor TCustomPipe.Destroy;
begin
  InternalClose(DISCONNECT_REASON_DESTROYING);
  DeleteCriticalSection(FCSDisconnect);
  inherited;
end;

function TCustomPipe._ExecuteReceiving(T: TCustomThread): Integer;
label
  finish;
var
  R: Boolean;
  Err: Integer;
  Buf: TBuffer;
  BR: Integer;
  Overlap: TOverlapped;
  //FEvRead: THandle;
begin
  //       
  // (   TCustomPipe, Self)
  //   ,     ,
  //      
  IncRef;
  
  Result := 0;
  
  //if FEvRead = 0 then FEvRead := CreateEvent(nil, False{ManualReset}, False, nil);
  
  ZeroMemory(@Overlap, SizeOf(TOverlapped));
  Overlap.hEvent := CreateEvent(nil, False{ManualReset}, False, nil);
  
  Buf := TBuffer.Create;
  Buf.Size := ReadBufferSize;
  
  repeat
    //ZeroMemory(@Overlap, SizeOf(TOverlapped));
    //Overlap.hEvent := FEvRead;
    
    BR := 0;
    
    R := ReadFile(FPipeHandle, Buf.Buffer^, ReadBufferSize, DWORD(BR), @Overlap);
    
    if not R then
    begin
      Err := GetLastError();
    
      if Err = ERROR_IO_PENDING then //     - 
      begin
        while True do
        begin
          BR := WaitForOverlappedResult(FPipeHandle, Overlap, 1000);
          
          if T.Stopped then
          begin
            CancelIo(FPipeHandle);
            goto finish;
          end;
          
          if BR = RESULT_ERROR then // 
          begin
            if AsyncEvents then
              _SyncDisconnect(nil)       
            else
              T.Synchronize(_SyncDisconnect, nil);
            goto finish;
          end;
          
          {if BR = 0 then // ?
          begin
            
          end;}
          
          if BR > 0 then Break; //   -   , BR = -1 = RESULT_TIMEOUT -   
        end;
      end
      else
      begin
        if AsyncEvents then
          _SyncDisconnect(nil)       
        else
          T.Synchronize(_SyncDisconnect, nil);
        goto finish;
      end;
    end;
    
    //  :
    if BR > 0 then
    begin
      Buf.Len := BR;
      if AsyncEvents then
        _SyncReceive(Buf)        
      else
        T.Synchronize(_SyncReceive, Buf);
    end;
    
  until T.Stopped;
 
  
finish:
  
  if not T.Stopped then
  begin
    FReadingThread := nil;
  end;
  
  CloseHandle(Overlap.hEvent);
  
  Buf.Free;
  
  DecRef;
end;

procedure TCustomPipe._SyncDisconnect(Param: Pointer);
begin
  InternalClose(DISCONNECT_REASON_REMOTE);
end;

procedure TCustomPipe._SyncReceive(Param: Pointer);
begin
  DataReceived(TBuffer(Param));
end;

procedure TCustomPipe.AfterClose(Reason: Integer);
begin
  
end;

procedure TCustomPipe.BeforeClose(Reason: Integer);
begin
  StopReceiving;
end;

procedure TCustomPipe.DataReceived(Buf: TBuffer);
begin

end;

procedure TCustomPipe.Disconnect;
begin
  InternalClose(DISCONNECT_REASON_MANUAL);
end;

procedure TCustomPipe.InternalClose(Reason: Integer);
begin
  EnterCriticalSection(FCSDisconnect);
  if FPipeHandle <> 0 then //     InternalClose-Destroy-InternalClose-...
  begin
    BeforeClose(Reason);
    Close;
    LeaveCriticalSection(FCSDisconnect);
    AfterClose(Reason);
  end
  else
    LeaveCriticalSection(FCSDisconnect);
end;

function TCustomPipe.Send(const S: string): Integer;
begin
  Result := Write(Pointer(S), Length(S));
end;

procedure TCustomPipe.StartReceiving;
begin
  if (FPipeHandle = 0) or
     (FReadingThread <> nil)
  then
    Exit;
    
  FReadingThread := TThread.Create;
  FReadingThread.OnExecute := _ExecuteReceiving;
  FReadingThread.AutoFree := True;
  FReadingThread.Resume;
end;

procedure TCustomPipe.StopReceiving;
begin
  if FReadingThread <> nil then
  begin
    FReadingThread.Stop;
    FReadingThread := nil;
  end;
end;

// ====== TCustomPipeClient ====== //

constructor TCustomPipeClient.Create;
begin
  inherited Create(0);
end;

function TCustomPipeClient.Connect(Name: string; Mode: Integer): Boolean;
label
  again;
var
  PipeH: THandle;
  Err: Integer;
begin
  if FPipeHandle <> 0 then
  begin
    Result := False;
    FErrorCode := ERROR_PIPE_CONNECTED;
    Exit;
  end;
  
  Err := 0;
  
again:
  PipeH := CreateFile(
    Pointer(Name),
    Mode,
    0,
    nil,
    OPEN_EXISTING,
    FILE_FLAG_OVERLAPPED,
    0
  );
  
  if PipeH = INVALID_HANDLE_VALUE then
  begin
    if Err = 0 then //  
    begin
      Err := GetLastError();
      
      // Pipe ,      ConnectNamedPipe() -
      //     
      if (Err = ERROR_PIPE_BUSY) and // ERROR_PIPE_BUSY = 231
         WaitNamedPipe(Pointer(Name), ConnectTimeout) //   - ERROR_SEM_TIMEOUT=121
      then
        goto again;
    end;
    
    Result := False;
    FErrorCode := Err;
  end
  else
  begin
    FPipeHandle := PipeH;
    Result := True;
    FErrorCode := 0;    
  end;
end;

// ====== TBasicPipeServer ====== //

constructor TBasicPipeServer.Create;
begin
  inherited;
  ReadBufferSize := DEFAULT_BUFFER_SIZE;
  WriteBufferSize := DEFAULT_BUFFER_SIZE;
  WriteTimeout := DEFAULT_WRITE_TIMEOUT;
end;

destructor TBasicPipeServer.Destroy;
begin
  StopListeninig;
  inherited;
end;

function TBasicPipeServer._ExecuteAccept(T: TCustomThread): Integer;
label
  finish;
var
  R: Boolean;
  Err: Integer;
  BR: Integer;
  Overlap: TOverlapped;
  PipeH: THandle;
begin
  Result := 0;
  
  ZeroMemory(@Overlap, SizeOf(TOverlapped));
  Overlap.hEvent := CreateEvent(nil, False{ManualReset}, False, nil);
  
  PipeH := FPipeHandle;
  
  
  while True do
  begin
    R := ConnectNamedPipe(PipeH, @Overlap);
    Err := GetLastError();
    
    R := R or (Err = ERROR_PIPE_CONNECTED);
    
    if T.Stopped then
    begin
      CancelIo(PipeH);
      goto finish;
    end;
    
    BR := 0;
    
    if not R then
    begin
      if Err = ERROR_IO_PENDING then
      begin
        while True do
        begin
          BR := WaitForOverlappedResult(PipeH, Overlap, 1000);
          
          if T.Stopped then
          begin
            CancelIo(PipeH);
            goto finish;
          end;
          
          if BR = RESULT_ERROR then // 
          begin
            Err := GetLastError();
            CancelIo(PipeH);
            Break;
          end;
          
          if BR <> RESULT_TIMEOUT then // 
          begin
            Break;
          end;
        end;
      end
      else // 
      begin
        BR := RESULT_ERROR;
      end;
    end;
    
    //  
    if BR >= 0 then
    begin
      if AsyncEvents then
        _SyncClientConnect(Pointer(PipeH))
      else
        T.Synchronize(_SyncClientConnect, Pointer(PipeH));
        
      PipeH := 0; //      ,  
      
      if T.Stopped then goto finish;
    end
    else // 
    begin
      if not T.Stopped then
      begin
        //       ListeninigFailed()
        FAcceptThread := nil;
        T.Stop;
        
        if AsyncEvents then
          _SyncListeningError(Pointer(Err))
        else
          T.Synchronize(_SyncListeningError, Pointer(Err));
      end;
        
      goto finish;
    end;
    
    //   
    PipeH := CreatePipeInstance;
    if PipeH = INVALID_HANDLE_VALUE then // 
    begin
      Err := GetLastError();
      if not T.Stopped then
      begin
        //       ListeninigFailed()
        FAcceptThread := nil;
        T.Stop;
        
        if AsyncEvents then
          _SyncListeningError(Pointer(Err))
        else
          T.Synchronize(_SyncListeningError, Pointer(Err));
      end;
      
      goto finish;
    end;
  end;
  
  
finish:
  
  if not T.Stopped then
  begin
    FAcceptThread := nil;
  end;
  
  CloseHandle(PipeH);
  CloseHandle(Overlap.hEvent);
end;

procedure TBasicPipeServer._SyncClientConnect(Param: Pointer);
begin
  ConnectionAccepted(THandle(Param));
end;

procedure TBasicPipeServer._SyncListeningError(Param: Pointer);
begin
  ListeninigFailed(Integer(Param));
end;

function TBasicPipeServer.CreatePipeInstance: THandle;
begin
  Result := CreateNamedPipe(
      Pointer(FPipeName), // lpName
      FPipeMode or FILE_FLAG_OVERLAPPED, // dwOpenMode
      PIPE_TYPE_BYTE or PIPE_READMODE_BYTE, // dwPipeMode (   PIPE_TYPE_MESSAGE or PIPE_READMODE_MESSAGE)
      PIPE_UNLIMITED_INSTANCES {255}, // nMaxInstances
      WriteBufferSize, // nOutBufferSize
      ReadBufferSize, // nInBufferSize
      0, // nDefaultTimeOut (   WaitNamedPipe)
      nil //lpSecurityAttributes
    );
end;

function TBasicPipeServer.GetIsListening: Boolean;
begin
  Result := (FAcceptThread <> nil);
end;

procedure TBasicPipeServer.ConnectionAccepted(PipeHandle: THandle);
begin
  CloseHandle(PipeHandle);
end;


procedure TBasicPipeServer.ListeninigFailed(ErrorCode: Integer);
begin

end;

function TBasicPipeServer.StartListeninig(Name: string; Mode: Integer): Integer;
begin
  if FAcceptThread <> nil then
  begin
    Result := ERROR_BUSY;
    Exit;
  end;

  FPipeName := Name;
  FPipeMode := Mode;
  
  FPipeHandle := CreatePipeInstance;
  if FPipeHandle = INVALID_HANDLE_VALUE then
  begin
    Result := GetLastError();
  end
  else
  begin
    Result := 0;
    FAcceptThread := TThread.Create;
    FAcceptThread.OnExecute := _ExecuteAccept;
    FAcceptThread.AutoFree := True;
    FAcceptThread.Resume;
  end;
end;

procedure TBasicPipeServer.StopListeninig;
begin
  if FAcceptThread <> nil then
  begin
    FAcceptThread.Stop;
    FAcceptThread := nil;
  end;
end;


// ====== TCustomPipeConnection ====== //

procedure TCustomPipeConnection.AfterClose(Reason: Integer);
begin
  if Server <> nil then
    Server.Disconnected(Self);
  
  DecRef; //   (   .)
end;

procedure TCustomPipeConnection.DataReceived(Buf: TBuffer);
begin
  if Server <> nil then
    Server.DataReceived(Self, Buf);
end;

// ====== TCustomPipeServer ====== //

constructor TCustomPipeServer.Create;
begin
  inherited;
  InitializeCriticalSection(FCSList);
  FConnList := NewList;
  FConnList.AddBy := 20;
end;

destructor TCustomPipeServer.Destroy;
begin
  StopServer;

  EnterCriticalSection(FCSList);
    FConnList.Free;
  LeaveCriticalSection(FCSList);
  
  DeleteCriticalSection(FCSList);
  inherited;
end;

procedure TCustomPipeServer.AbortSend(Conn: TCustomPipeConnection);
begin
  Conn := HoldConnection(Conn);
  if Conn <> nil then
  begin
    Conn.AbortWrite;
    Conn.DecRef;
  end;
end;

procedure TCustomPipeServer.AbortSendAll;
var 
  C: TCustomPipeConnection;
  I: Integer;
begin
  FAbortSendAll := False;
  EnterCriticalSection(FCSList);
    //if Assigned(FConnList) then
    //begin
      for I := 0 to FConnList.Count - 1 do
      begin
        C := FConnList.Items[I];
        C.AbortWrite;
      end;
    //end;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomPipeServer.AfterNewConnection(Conn: TCustomPipeConnection);
begin

end;

function TCustomPipeServer.BeforeNewConnection(PipeHandle: THandle): TCustomPipeConnection;
begin
  Result := nil;
end;

procedure TCustomPipeServer.Close(Conn: TCustomPipeConnection);
begin
  Conn := HoldConnection(Conn);
  if Conn <> nil then
  begin
    Conn.Disconnect;
    Conn.DecRef;
  end;
end;

procedure TCustomPipeServer.CloseAll;
var 
  C: TCustomPipeConnection;
  I: Integer;
begin
  EnterCriticalSection(FCSList);
    for I := 0 to FConnList.Count - 1 do
    begin
      C := FConnList.Items[I];
      //C.Server := nil; //   Server.ConnectionDelete()  C.InternalClose()  :  Server    AfterClose()
      FConnList.Items[I] := nil; //       Server.ConnectionDelete()  C.InternalClose()
      C.InternalClose(DISCONNECT_REASON_MANUAL); //    TCustomPipeConnection.AfterClose
    end;
    FConnList.Clear;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomPipeServer.ConnectionAccepted(PipeHandle: THandle);
var
  Conn: TCustomPipeConnection;
begin
  Conn := BeforeNewConnection(PipeHandle);
  if Conn = nil then
  begin
    inherited;
    Exit;
  end;
  Conn.AsyncEvents := AsyncEvents;
  Conn.ReadBufferSize := ReadBufferSize;
  Conn.WriteBufferSize := WriteBufferSize;
  Conn.WriteTimeout := WriteTimeout;
  
  ConnectionAdd(Conn);
  AfterNewConnection(Conn);
end;


procedure TCustomPipeServer.ConnectionAdd(Conn: TCustomPipeConnection);
begin
  EnterCriticalSection(FCSList);
    if FConnList.IndexOf(Conn) = -1 then
    begin
      Conn.Server := Self;
      FConnList.Add(Conn);
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomPipeServer.ConnectionDelete(Conn: TCustomPipeConnection);
var
  I: Integer;
begin
  EnterCriticalSection(FCSList);
    I := FConnList.IndexOf(Conn);
    if I <> -1 then
    begin
      FConnList.Delete(I);
      Conn.Server := nil;
      //Conn.DecRef; //   (   .)
    end;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomPipeServer.DataReceived(Conn: TCustomPipeConnection; Buf: TBuffer);
begin

end;

procedure TCustomPipeServer.Disconnected(Conn: TCustomPipeConnection);
begin
  ConnectionDelete(Conn);
end;

procedure TCustomPipeServer.GetConnections(OutConn: PList);
var
  I: Integer;
begin
  if OutConn = nil then Exit;
  EnterCriticalSection(FCSList);
    for I := 0 to FConnList.Count - 1 do
    begin
      OutConn.Add(FConnList.Items[I]);
    end;
  LeaveCriticalSection(FCSList);
end;

function TCustomPipeServer.GetCount: Integer;
begin
  Result := FConnList.Count;
end;

function TCustomPipeServer.HoldConnection(Conn: TCustomPipeConnection): TCustomPipeConnection;
begin
  EnterCriticalSection(FCSList);
    if FConnList.IndexOf(Conn) > -1 then
    begin
      Result := Conn;
      Result.IncRef;
    end
    else
      Result := nil;
  LeaveCriticalSection(FCSList);
end;

function TCustomPipeServer.HoldConnectionIdx(I: Integer): TCustomPipeConnection;
begin
  EnterCriticalSection(FCSList);
    if (I >= 0) and (I < FConnList.Count) then
    begin
      Result := FConnList.Items[I];
      Result.IncRef;
    end
    else
      Result := nil;
  LeaveCriticalSection(FCSList);
end;

procedure TCustomPipeServer.Send(Conn: TCustomPipeConnection; S: string);
begin
  Conn := HoldConnection(Conn);
  if Conn <> nil then
  begin
    if Conn.Send(S) >= 0 then
      SendResult(Conn, 0)
    else
      SendResult(Conn, Conn.ErrorCode);
    Conn.DecRef;
  end;
end;

procedure TCustomPipeServer.SendAll(S: string);
var
  I: Integer;
  Conn: TCustomPipeConnection;
begin
  if FSendAllStarted then Exit;
  
  FSendAllStarted := True;
  FAbortSendAll := False;
  
  
  I := 0;
  Conn := HoldConnectionIdx(I);
  while (Conn <> nil) do
  begin
    if Conn.Send(S) >= 0 then
      SendResult(Conn, 0)
    else
      SendResult(Conn, Conn.ErrorCode);
      
    Conn.DecRef;
    
    if FAbortSendAll then Break;
    
    Inc(I);
    Conn := HoldConnectionIdx(I);
  end;
  
  FSendAllStarted := False;
end;

procedure TCustomPipeServer.SendResult(Conn: TCustomPipeConnection; ResultCode: Integer);
begin
  
end;

procedure TCustomPipeServer.StopServer;
begin
  StopListeninig;
  CloseAll;
end;

// ====== TPipeClient ====== //

procedure TPipeClient.AfterClose(Reason: Integer);
begin
  if Assigned(OnDisconnect) then OnDisconnect(Self, Reason);
end;

procedure TPipeClient.DataReceived(Buf: TBuffer);
begin
  if Assigned(OnReceive) then OnReceive(Self, Buf);
end;


// ====== TPipeServer ====== //

procedure TPipeServer.DataReceived(Conn: TCustomPipeConnection; Buf: TBuffer);
begin
  if Assigned(OnReceive) then OnReceive(Conn, Buf);
end;

procedure TPipeServer.Disconnected(Conn: TCustomPipeConnection);
begin
  // Conn    TCustomPipeConnection.AfterClose()
  inherited; //    
  if Assigned(OnClientDisconnect) then OnClientDisconnect(Conn);
end;

procedure TPipeServer.AfterNewConnection(Conn: TCustomPipeConnection);
begin
  if Assigned(OnClientConnect) then 
    OnClientConnect(Conn);
    
  if PipeMode <> PIPE_ACCESS_OUTBOUND then
    Conn.StartReceiving;
end;

function TPipeServer.BeforeNewConnection(PipeHandle: THandle): TCustomPipeConnection;
begin
  Result := TCustomPipeConnection.Create(PipeHandle);
end;

procedure TPipeServer.SendResult(Conn: TCustomPipeConnection; ResultCode: Integer);
begin
  if ResultCode = 0 then
  begin
    if Assigned(OnSend) then OnSend(Conn);
  end
  else
  begin
    if Assigned(OnErrorSend) then OnErrorSend(Conn, ResultCode);
  end;
end;

end.
