unit hiPipeClient;

interface

uses
  Windows, Kol, Share, NSPipes;


type
  
  ThiPipeClient = class
    private
      FPipeClient: TPipeClient;
      
      procedure SetAsyncEvents(Value: Integer);
      procedure SetConnectTimeout(Value: Integer);
      procedure SetReadBufSize(Value: Integer);
      procedure SetSendTimeout(Value: Integer);
      procedure SetWriteBufSize(Value: Integer);
      
      procedure _OnDisconnect(Sender: TCustomPipe; Reason: Integer);
      procedure _OnReceive(Sender: TCustomPipe; Buf: TBuffer);
      
    public
      _prop_PipeName: string;
      _prop_Mode: Integer;
      
      _data_Data: THI_Event;
      _data_PipeName: THI_Event;
      
      _event_onConnect: THI_Event;
      _event_onDisconnect: THI_Event;
      _event_onErrorConnect: THI_Event;
      _event_onErrorSend: THI_Event;
      _event_onReceive: THI_Event;
      _event_onSend: THI_Event;
      
      property _prop_AsyncEvents: Integer write SetAsyncEvents;
      property _prop_ConnectTimeout: Integer write SetConnectTimeout;
      property _prop_ReadBufSize: Integer write SetReadBufSize;
      property _prop_SendTimeout: Integer write SetSendTimeout;
      property _prop_WriteBufSize: Integer write SetWriteBufSize;
      
      constructor Create;
      destructor Destroy; override;
      
      procedure _work_doAbortSend(var _Data: TData; Index: Word);
      procedure _work_doConnect(var _Data: TData; Index: Word);
      procedure _work_doDisconnect(var _Data: TData; Index: Word);
      procedure _work_doSend(var _Data: TData; Index: Word);
  end;
  

implementation

constructor ThiPipeClient.Create;
begin
  inherited;
  FPipeClient := TPipeClient.Create;
  FPipeClient.OnDisconnect := _OnDisconnect;
  FPipeClient.OnReceive := _OnReceive;
end;

destructor ThiPipeClient.Destroy;
begin
  FPipeClient.OnDisconnect := nil;
  FPipeClient.Disconnect;
  FPipeClient.DecRef;
  inherited;
end;

procedure ThiPipeClient._OnDisconnect(Sender: TCustomPipe; Reason: Integer);
begin
  _hi_OnEvent(_event_onDisconnect, Reason);
end;

procedure ThiPipeClient._OnReceive(Sender: TCustomPipe; Buf: TBuffer);
begin
  _hi_OnEvent(_event_onReceive, Buf.GetString);
end;

procedure ThiPipeClient.SetAsyncEvents(Value: Integer);
begin
  FPipeClient.AsyncEvents := Value <> 0;
end;

procedure ThiPipeClient.SetConnectTimeout(Value: Integer);
begin
  FPipeClient.ConnectTimeout := Value;
end;

procedure ThiPipeClient.SetReadBufSize(Value: Integer);
begin
  FPipeClient.ReadBufferSize := Value;
end;

procedure ThiPipeClient.SetSendTimeout(Value: Integer);
begin
  FPipeClient.WriteTimeout := Value;
end;

procedure ThiPipeClient.SetWriteBufSize(Value: Integer);
begin
  FPipeClient.WriteBufferSize := Value;
end;

procedure ThiPipeClient._work_doAbortSend(var _Data: TData; Index: Word);
begin
  FPipeClient.AbortWrite;
end;

procedure ThiPipeClient._work_doConnect(var _Data: TData; Index: Word);
var
  N: string;
  M: DWORD;
begin
  //  Connect()   
  //if FPipeHandle <> 0 then Exit;
  
  N := ReadString(_Data, _data_PipeName, _prop_PipeName);
  if N = '' then Exit;
  
  N := '\\.\pipe\' + N;
  
  case _prop_Mode of 
    0: M := GENERIC_READ;
    1: M := GENERIC_WRITE;
    else
      M := GENERIC_READ or GENERIC_WRITE;
  end;
  
  if not FPipeClient.Connect(N, M) then
    _hi_CreateEvent(_Data, @_event_onErrorConnect, FPipeClient.ErrorCode)
  else
  begin
    _hi_CreateEvent(_Data, @_event_onConnect);
    if _prop_Mode <> 1 then FPipeClient.StartReceiving;
  end;
end;

procedure ThiPipeClient._work_doDisconnect(var _Data: TData; Index: Word);
begin
  FPipeClient.Disconnect;
end;

procedure ThiPipeClient._work_doSend(var _Data: TData; Index: Word);
var
  R: Integer;
begin
  R := FPipeClient.Send(ReadString(_Data, _data_Data, ''));
  if R >= 0 then
    _hi_CreateEvent(_Data, @_event_onSend)
  else
    if FPipeClient.ErrorCode <> ERROR_OPERATION_ABORTED then
      _hi_CreateEvent(_Data, @_event_onErrorSend, FPipeClient.ErrorCode);
end;

end.
