unit hiPipeServer;

interface

uses
  Windows, Kol, Share, NSPipes, NSThreads;


type
    
  ThiPipeServer = class(TCustomPipeServer)
    private
      FCSClient: TRTLCriticalSection;
      FNoSync: Boolean;
      
      procedure SetAsyncEvents(Value: Integer);
      procedure SetSendTimeout(Value: Integer);
      
    protected
      FCurClientID: Integer;
      
      procedure AfterNewConnection(Conn: TCustomPipeConnection); override;
      function BeforeNewConnection(PipeHandle: THandle): TCustomPipeConnection; override;
      //procedure ListeninigFailed(ErrorCode: Integer); override;
      procedure SendResult(Conn: TCustomPipeConnection; ResultCode: Integer); override;
      
      procedure DataReceived(Conn: TCustomPipeConnection; Buf: TBuffer); override;
      procedure Disconnected(Conn: TCustomPipeConnection); override;
      
    public
      _prop_PipeName: string;
      _prop_Mode: Integer;
      
      _data_ClientID: THI_Event;
      _data_Data: THI_Event;
      _data_PipeName: THI_Event;
      
      
      _event_onClientConnect: THI_Event;
      _event_onClientDisconnect: THI_Event;
      _event_onEnumClients: THI_Event;
      _event_onErrorSend: THI_Event;
      _event_onErrorStart: THI_Event;
      _event_onReceive: THI_Event;
      _event_onSend: THI_Event;
      _event_onStartServer:THI_Event;
      
      property _prop_AsyncEvents: Integer write SetAsyncEvents;
      property _prop_ReadBufSize: Integer write ReadBufferSize;
      property _prop_SendTimeout: Integer write SetSendTimeout;
      property _prop_WriteBufSize: Integer write WriteBufferSize;
      
      constructor Create;
      destructor Destroy; override;
      
      procedure _work_doAbortSend(var _Data: TData; Index: Word);
      procedure _work_doAbortSendAll(var _Data: TData; Index: Word);
      procedure _work_doClose(var _Data: TData; Index: Word);
      procedure _work_doCloseAll(var _Data: TData; Index: Word);
      procedure _work_doEnumClients(var _Data: TData; Index: Word);
      procedure _work_doSend(var _Data: TData; Index: Word);
      procedure _work_doSendAll(var _Data: TData; Index: Word);
      procedure _work_doStartServer(var _Data: TData; Index: Word);
      procedure _work_doStopListening(var _Data: TData; Index: Word);
      procedure _work_doStopServer(var _Data: TData; Index: Word);
      
      procedure _var_Count(var _Data: TData; Index: Word);
      procedure _var_CurClientID(var _Data: TData; Index: Word);
      procedure _var_Listening(var _Data: TData; Index: Word);
  end;


implementation


// ====== ThiPipeServer ====== //

constructor ThiPipeServer.Create;
begin
  inherited;
  InitializeCriticalSection(FCSClient);
end;

destructor ThiPipeServer.Destroy;
begin
  _event_onClientDisconnect.Event := nil; //   
  inherited;
  DeleteCriticalSection(FCSClient);
end;

procedure ThiPipeServer.AfterNewConnection(Conn: TCustomPipeConnection);
begin
  if FNoSync then EnterCriticalSection(FCSClient);
  
  FCurClientID := Integer(Conn);
  _hi_OnEvent(_event_onClientConnect, Integer(Conn));
  
  if PipeMode <> PIPE_ACCESS_OUTBOUND then
    Conn.StartReceiving;

  if FNoSync then LeaveCriticalSection(FCSClient);
end;

function ThiPipeServer.BeforeNewConnection(PipeHandle: THandle): TCustomPipeConnection;
begin
  Result := TCustomPipeConnection.Create(PipeHandle);
end;

procedure ThiPipeServer.DataReceived(Conn: TCustomPipeConnection; Buf: TBuffer);
begin
  if FNoSync then EnterCriticalSection(FCSClient);
  
  FCurClientID := Integer(Conn);
  _hi_OnEvent(_event_onReceive, Buf.GetString);
  
  if FNoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiPipeServer.Disconnected(Conn: TCustomPipeConnection);
begin
  inherited;
  
  if FNoSync then EnterCriticalSection(FCSClient);
  
  FCurClientID := Integer(Conn);
  _hi_OnEvent(_event_onClientDisconnect, Integer(Conn));
  
  if FNoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiPipeServer.SendResult(Conn: TCustomPipeConnection; ResultCode: Integer);
begin
  if FNoSync then EnterCriticalSection(FCSClient);
  
  FCurClientID := Integer(Conn);
  
  if ResultCode = 0 then
    _hi_OnEvent(_event_onSend, Integer(Conn))
  else
    if ResultCode <> ERROR_OPERATION_ABORTED then
      _hi_OnEvent(_event_onErrorSend, ResultCode);
      
  if FNoSync then LeaveCriticalSection(FCSClient);
end;

procedure ThiPipeServer.SetAsyncEvents(Value: Integer);
begin
  AsyncEvents := Value <> 0;
  FNoSync := (not SyncAvailable) or AsyncEvents;
end;

procedure ThiPipeServer.SetSendTimeout(Value: Integer);
begin
  if Value <= 0 then Value := Integer(INFINITE);
  
  WriteTimeout := Value;
end;

procedure ThiPipeServer._work_doAbortSend(var _Data: TData; Index: Word);
begin
  AbortSend(TCustomPipeConnection(ReadInteger(_Data, _data_ClientID, 0)));
end;

procedure ThiPipeServer._work_doAbortSendAll(var _Data: TData; Index: Word);
begin
  AbortSendAll;
end;

procedure ThiPipeServer._work_doClose(var _Data: TData; Index: Word);
begin
  Close(TCustomPipeConnection(ReadInteger(_Data, _data_ClientID, 0)));
end;

procedure ThiPipeServer._work_doCloseAll(var _Data: TData; Index: Word);
begin
  CloseAll;
end;

procedure ThiPipeServer._work_doEnumClients(var _Data: TData; Index: Word);
var
  Lst: PList;
  I: Integer;
begin
  Lst := NewList;
  GetConnections(Lst);
    for I := 0 to Lst.Count - 1 do
    begin
      if FNoSync then EnterCriticalSection(FCSClient);
        FCurClientID := Integer(Lst.Items[I]);
        _hi_OnEvent(_event_onEnumClients, FCurClientID);
      if FNoSync then LeaveCriticalSection(FCSClient);
    end;
  Lst.Free;
end;

procedure ThiPipeServer._work_doSend(var _Data: TData; Index: Word);
var
  S: string;
  I: Integer;
begin
  S := ReadString(_Data, _data_Data, '');
  I := ReadInteger(_Data, _data_ClientID, 0);
  Send(TCustomPipeConnection(I), S);
end;

procedure ThiPipeServer._work_doSendAll(var _Data: TData; Index: Word);
begin
  SendAll(ReadString(_Data, _data_Data, ''));
end;

procedure ThiPipeServer._work_doStartServer(var _Data: TData; Index: Word);
var
  Err: Integer;
  N: string;
  M: Integer;
begin
  N := ReadString(_Data, _data_PipeName, _prop_PipeName);
  if N <> '' then
  begin
    N := '\\.\pipe\' + N;

    {case _prop_Mode of 
      0: M := PIPE_ACCESS_INBOUND;
      1: M := PIPE_ACCESS_OUTBOUND;
      else
        M := PIPE_ACCESS_DUPLEX;
    end;}
    M := _prop_Mode + 1;
    
    Err := StartListeninig(N, M);
    if Err <> 0 then
      _hi_CreateEvent(_Data, @_event_onErrorStart, Err)
    else
      _hi_CreateEvent(_Data, @_event_onStartServer);
  end;
end;

procedure ThiPipeServer._work_doStopListening(var _Data: TData; Index: Word);
begin
  StopListeninig;
end;

procedure ThiPipeServer._work_doStopServer(var _Data: TData; Index: Word);
begin
  StopServer;
end;

procedure ThiPipeServer._var_Count;
begin
  dtInteger(_Data, Count);
end;

procedure ThiPipeServer._var_CurClientID;
begin
  dtInteger(_Data, FCurClientID);
end;

procedure ThiPipeServer._var_Listening;
begin
  dtInteger(_Data, Ord(IsListening));
end;

end.
